package ch.unige.solidify.util;

import java.io.IOException;
import java.nio.file.Paths;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.ClassPathResource;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyBagItCheckingException;
import ch.unige.solidify.exception.SolidifyBagItException;
import ch.unige.solidify.exception.SolidifyRuntimeException;

import gov.loc.repository.bagit.conformance.BagProfileChecker;
import gov.loc.repository.bagit.domain.Bag;
import gov.loc.repository.bagit.exceptions.CorruptChecksumException;
import gov.loc.repository.bagit.exceptions.FileNotInPayloadDirectoryException;
import gov.loc.repository.bagit.exceptions.InvalidBagitFileFormatException;
import gov.loc.repository.bagit.exceptions.InvalidPayloadOxumException;
import gov.loc.repository.bagit.exceptions.MaliciousPathException;
import gov.loc.repository.bagit.exceptions.MissingBagitFileException;
import gov.loc.repository.bagit.exceptions.MissingPayloadDirectoryException;
import gov.loc.repository.bagit.exceptions.MissingPayloadManifestException;
import gov.loc.repository.bagit.exceptions.UnparsableVersionException;
import gov.loc.repository.bagit.exceptions.UnsupportedAlgorithmException;
import gov.loc.repository.bagit.exceptions.VerificationException;
import gov.loc.repository.bagit.exceptions.conformance.BagitVersionIsNotAcceptableException;
import gov.loc.repository.bagit.exceptions.conformance.FetchFileNotAllowedException;
import gov.loc.repository.bagit.exceptions.conformance.MetatdataValueIsNotAcceptableException;
import gov.loc.repository.bagit.exceptions.conformance.MetatdataValueIsNotRepeatableException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredManifestNotPresentException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredMetadataFieldNotPresentException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredTagFileNotPresentException;
import gov.loc.repository.bagit.reader.BagReader;
import gov.loc.repository.bagit.verify.BagVerifier;

public class BagItTool {
  private static final Logger log = LoggerFactory.getLogger(BagItTool.class);

  // Check conformance of BagIt package
  public static void check(String bagItProfile, String bagItPackage) {
    final Bag bag = read(bagItPackage);

    final boolean ignoreHiddenFiles = true;
    final ExecutorService exeService = new ThreadPoolExecutor(0, 3000, 60L, TimeUnit.SECONDS, new LinkedBlockingQueue<>());
    try (BagVerifier bagVerifier = new BagVerifier(exeService)) {
      bagVerifier.isValid(bag, ignoreHiddenFiles);
      BagProfileChecker.bagConformsToProfile(new ClassPathResource(bagItProfile).getInputStream(), bag);
    } catch (final InterruptedException e) {
      // Restore interrupted state...
      Thread.currentThread().interrupt();
      throw new SolidifyRuntimeException("BagItTool#check was interrupted!", e);
    } catch (IOException | MissingPayloadManifestException | MissingBagitFileException
            | MissingPayloadDirectoryException | FileNotInPayloadDirectoryException
            | MaliciousPathException | UnsupportedAlgorithmException
            | InvalidBagitFileFormatException | CorruptChecksumException | VerificationException
            | FetchFileNotAllowedException | RequiredMetadataFieldNotPresentException
            | MetatdataValueIsNotAcceptableException | RequiredManifestNotPresentException
            | BagitVersionIsNotAcceptableException | RequiredTagFileNotPresentException
            | MetatdataValueIsNotRepeatableException e) {
      throw new SolidifyBagItCheckingException("Error while checking BagIt package " + bag, e);
    }
  }

  public static void quickCheck(String bagItProfile, String bagItPackage) {
    final Bag bag = read(bagItPackage);
    try {
      if (BagVerifier.canQuickVerify(bag)) {
        BagVerifier.quicklyVerify(bag);
      } else {
        log.warn("Cannot proceed a quick check for BagIt package {}", bag);
      }
      BagProfileChecker.bagConformsToProfile(new ClassPathResource(bagItProfile).getInputStream(), bag);
    } catch (IOException | FetchFileNotAllowedException
            | RequiredMetadataFieldNotPresentException | MetatdataValueIsNotAcceptableException
            | RequiredManifestNotPresentException | BagitVersionIsNotAcceptableException
            | RequiredTagFileNotPresentException | InvalidPayloadOxumException
            | MetatdataValueIsNotRepeatableException e) {
      throw new SolidifyBagItCheckingException("Error while quick checking BagIt package " + bag, e);
    }
  }

  // Read BagIt package
  public static Bag read(String bagItPackage) {
    Bag bag;
    final BagReader bagReader = new BagReader();
    try {
      bag = bagReader.read(Paths.get(bagItPackage));
    } catch (IOException | UnparsableVersionException | MaliciousPathException
            | UnsupportedAlgorithmException | InvalidBagitFileFormatException e) {
      throw new SolidifyBagItException("Error reading BagIt package", e);
    }
    return bag;
  }

  /**
   * This constructor should not be called
   */
  private BagItTool() {
    throw new IllegalStateException(SolidifyConstants.TOOL_CLASS);
  }

}
