package ch.unige.solidify.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathFactory;

import org.json.JSONObject;
import org.json.XML;
import org.springframework.util.xml.SimpleNamespaceContext;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyProcessingException;

public class XMLTool {
  public static final String SCHEMA_FULL_VALIDATION = "http://apache.org/xml/features/validation/schema-full-checking";

  public static final String SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
  public static final String SCHEMA_PREFIX = "{XML Schema} ";
  public static final String SCHEMA_SOURCE = "http://java.sun.com/xml/jaxp/properties/schemaSource";

  public static final String SCHEMA_VALIDATION = "http://apache.org/xml/features/validation/schema";
  public static final String XML_PREFIX = "{XML Document} ";
  public static final String XML_VALIDATION = "http://xml.org/sax/features/validation";
  public static final String XSL_PREFIX = "{XSL Transformation} ";

  public static XPath buildXPath(Map<String, String> namespaces) {

    final XPathFactory xpathFactory = XPathFactory.newInstance();
    final XPath xpath = xpathFactory.newXPath();

    if (namespaces != null && !namespaces.isEmpty()) {
      final SimpleNamespaceContext nsc = new SimpleNamespaceContext();
      for (final Map.Entry<String, String> namespace : namespaces.entrySet()) {
        nsc.bindNamespaceUri(namespace.getKey(), namespace.getValue());
      }
      xpath.setNamespaceContext(nsc);
    }

    return xpath;
  }

  public static Path getOutputFile(Path xml) {
    return xml.getParent().resolve(xml.getFileName().toString()
            .replaceFirst(SolidifyConstants.XML_EXT, SolidifyConstants.XML_OUTPUT_EXT));
  }

  public static Document parseXML(String xml)
          throws ParserConfigurationException, SAXException, IOException {

    final DocumentBuilderFactory factory = getDocumentBuilderFactory();
    factory.setNamespaceAware(true);
    final DocumentBuilder builder = factory.newDocumentBuilder();
    return builder.parse(new InputSource(new StringReader(xml)));
  }

  public static void transform(Path xml, Path xsl) {
    transform(xml, xsl, getOutputFile(xml).toString());
  }

  public static void transform(Path xml, Path xsl, String newXml) {
    try {
      final StreamSource xmlSource = new StreamSource(Files.newInputStream(xml));
      final StreamSource xslSource = new StreamSource(Files.newInputStream(xsl));
      transform(xmlSource, xslSource, newXml);
    } catch (TransformerException | IOException e) {
      throw new SolidifyProcessingException(logTransformationError(xml.toString(), xsl.toString()), e);
    }
  }

  public static void transform(Path xml, String xsl) {
    transform(xml, xsl, getOutputFile(xml).toString());
  }

  public static void transform(Path xml, String xsl, String newXml) {
    try {
      final StreamSource xmlSource = new StreamSource(Files.newInputStream(xml));
      final StreamSource xslSource = new StreamSource(new StringReader(xsl));
      transform(xmlSource, xslSource, newXml);
    } catch (TransformerException | IOException e) {
      throw new SolidifyProcessingException(logTransformationError(xml.toString(), xsl), e);
    }
  }

  public static String transform(String xml, String xsl) {
    final StreamSource xmlSource = new StreamSource(new StringReader(xml));
    final StreamSource xslSource = new StreamSource(new StringReader(xsl));

    try {
      final TransformerFactory transformerFactory = TransformerFactory.newInstance();
      transformerFactory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
      final Transformer transformer = transformerFactory.newTransformer(xslSource);
      final StringWriter writer = new StringWriter();
      final StreamResult result = new StreamResult(writer);
      transformer.transform(xmlSource, result);
      return writer.toString();
    } catch (final TransformerException e) {
      throw new SolidifyProcessingException(logTransformationError(xml, xsl), e);
    }
  }

  public static String transform(String xml, String xsl, Map<String, String> parameters) {
    final StreamSource xmlSource = new StreamSource(new StringReader(xml));
    final StreamSource xslSource = new StreamSource(new StringReader(xsl));

    try {
      final TransformerFactory transformerFactory = TransformerFactory.newInstance();
      transformerFactory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
      final Transformer transformer = transformerFactory.newTransformer(xslSource);
      // Parameters
      for (final Map.Entry<String, String> entry : parameters.entrySet()) {
        transformer.setParameter(entry.getKey(), entry.getValue());
      }
      final StringWriter writer = new StringWriter();
      final StreamResult result = new StreamResult(writer);
      transformer.transform(xmlSource, result);
      return writer.toString();
    } catch (final TransformerException e) {
      throw new SolidifyProcessingException(logTransformationError(xml, xsl), e);
    }
  }

  public static void validate(List<URL> xsds, Path xml) {
    validate(xsds, xml, true);
  }

  public static void validate(List<URL> xsds, Path xml, boolean domParser) {

    final SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
    Source[] sources;
    Schema schema;

    // Schema factory
    try {
      sources = getSchemaSources(xsds);
      schemaFactory.setFeature(SCHEMA_FULL_VALIDATION, true);
      schemaFactory.setResourceResolver(new SchemaResourceResolver());
      schema = schemaFactory.newSchema(sources);
    } catch (IOException | SAXException e) {
      throw new SolidifyCheckingException(logValidationError(xml.toString(), null), e);
    }

    try {
      if (domParser) {
        parseWithDom(schema, xml);
      } else {
        parseWithSax(schema, xml);
      }
    } catch (ParserConfigurationException | SAXException | IOException e) {
      throw new SolidifyCheckingException("Error while parsing file " + xml + " with schema " + schema, e);
    }
  }

  public static void validate(String xsd, String xml) {

    DocumentBuilder parser;
    Document doc;

    try {
      // parse an XML document into a DOM tree
      parser = getDocumentBuilderFactory().newDocumentBuilder();
      parser.setErrorHandler(new XMLError());

      doc = parser.parse(new InputSource(new StringReader(xml)));

      // create a SchemaFactory capable of understanding WXS schemas
      final SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

      // load a WXS schema, represented by a Schema instance
      final Schema schema = factory.newSchema(new SAXSource(new InputSource(new StringReader(xsd))));

      // create a Validator instance, which can be used to validate an
      // instance document
      final Validator validator = schema.newValidator();
      // Disable external resources
      validator.setProperty(XMLConstants.ACCESS_EXTERNAL_DTD, "");
      validator.setProperty(XMLConstants.ACCESS_EXTERNAL_SCHEMA, "");

      // validate the DOM tree
      validator.validate(new DOMSource(doc));

    } catch (ParserConfigurationException | SAXException | IOException e) {
      throw new SolidifyCheckingException(logValidationError(xml, xsd), e);
    }
  }

  public static void wellformed(Path xml) {

    DocumentBuilderFactory factory;
    try {
      factory = getDocumentBuilderFactory();
      factory.setValidating(false);
      factory.setNamespaceAware(true);
    } catch (final ParserConfigurationException e) {
      throw new SolidifyCheckingException("Error in DocumentBuilderFactory configuration", e);
    }

    DocumentBuilder parser;
    try (InputStream inputStream = Files.newInputStream(xml)) {
      parser = factory.newDocumentBuilder();
      parser.setErrorHandler(new XMLError());
      parser.parse(new InputSource(new InputStreamReader(inputStream, StandardCharsets.UTF_8)));
    } catch (ParserConfigurationException | SAXException | IOException e) {
      throw new SolidifyCheckingException("Error while checking if file " + xml + " is well formed", e);
    }
  }

  public static JSONObject xml2Json(String xml) {
    return XML.toJSONObject(xml);
  }

  public static void xml2Stream(Marshaller m, OutputStream out, Object obj)
          throws JAXBException, TransformerFactoryConfigurationError, TransformerException {
    final DOMResult domResult = new DOMResult();

    m.marshal(obj, domResult);

    final TransformerFactory transformerFactory = TransformerFactory.newInstance();
    transformerFactory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
    final Transformer transformer = transformerFactory.newTransformer();
    transformer.setOutputProperty(OutputKeys.INDENT, "yes");
    transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
    transformer.transform(new DOMSource(domResult.getNode()), new StreamResult(out));
  }

  private static DocumentBuilderFactory getDocumentBuilderFactory() throws ParserConfigurationException {
    final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    factory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
    return factory;
  }

  private static Source[] getSchemaSources(List<URL> xsds) throws IOException {
    final Source[] sources = new Source[xsds.size()];
    int i = 0;
    for (final URL xsd : xsds) {
      sources[i] = new StreamSource(xsd.openStream());
      i++;
    }
    return sources;
  }

  private static String logTransformationError(String xml, String xsl) {
    final StringBuilder builder = new StringBuilder();
    builder.append("Error while transforming file " + xml);
    builder.append(" with " + xsl);
    return builder.toString();
  }

  private static String logValidationError(String xml, String xsd) {
    final StringBuilder builder = new StringBuilder();
    builder.append("Error while validating file " + xml);
    if (xsd != null) {
      builder.append(" with " + xsd);
    }
    return builder.toString();
  }

  private static void parseWithDom(Schema schema, Path xml)
          throws ParserConfigurationException, SAXException, IOException {
    final DocumentBuilderFactory factory = getDocumentBuilderFactory();
    // Doc Factory
    factory.setSchema(schema);
    factory.setValidating(false);
    factory.setNamespaceAware(true);

    // Parse document
    final DocumentBuilder parser = factory.newDocumentBuilder();
    parser.setErrorHandler(new XMLError());
    parser.parse(Files.newInputStream(xml));
  }

  private static void parseWithSax(Schema schema, Path xml)
          throws ParserConfigurationException, SAXException, IOException {
    final SAXParserFactory factory = SAXParserFactory.newInstance();
    // Disable external resources
    factory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
    // Doc Factory
    factory.setSchema(schema);
    factory.setValidating(false);
    factory.setNamespaceAware(true);

    // Parse document
    final XMLReader reader = factory.newSAXParser().getXMLReader();
    reader.setErrorHandler(new XMLError());
    reader.parse(new InputSource(Files.newInputStream(xml)));
  }

  private static void transform(StreamSource xmlSource, StreamSource xslSource, String newXml)
          throws UnsupportedEncodingException, TransformerException {
    final TransformerFactory transformerFactory = TransformerFactory.newInstance();
    transformerFactory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
    final Transformer transformer = transformerFactory.newTransformer(xslSource);

    final StreamResult xmlResult = new StreamResult(new File(newXml));
    xmlResult.setSystemId(java.net.URLDecoder.decode(xmlResult.getSystemId(), "UTF-8"));
    transformer.transform(xmlSource, xmlResult);
  }

  /**
   * This constructor should not be called
   */
  private XMLTool() {
    throw new IllegalStateException(SolidifyConstants.TOOL_CLASS);
  }

}
